/*
 * File:		lan91c11.c
 * Purpose:		Device driver for the SMSC LAN91C111
 *
 * Notes:		
 *
 */

#include "src/include/dbug.h"
#include "src/uif/net/net.h"
#include "src/include/dev/lan91c111.h"

/********************************************************************/

static int	lan91c111_reset (NIF *);
static void	lan91c111_start (NIF *);
static void	lan91c111_stop (NIF *);
static int	lan91c111_send (NIF *, HWA_ADDR_P, HWA_ADDR_P, uint16, NBUF*);
static void	lan91c111_receive (void *, NIF *);
static uint16 lan91c111_read_phy_reg(uint8, uint8);
static void	lan91c111_write_phy_reg(uint8, uint8, uint16);
static void	lan91c111_phy_configure(void);
static void	lan91c111_phy_shutdown(void);



/********************************************************************/
int
lan91c111_init (NIF *nif, int vector)
{
	nif_init(nif, "lan91c111");
	nif->vector = vector;

	nif->hwa_size = 6;
	nif->mtu = 1500;
	nif->reset = lan91c111_reset;
	nif->start = lan91c111_start;
	nif->stop = lan91c111_stop;
	nif->send = lan91c111_send;
	nif->receive = (void *)&lan91c111_receive;
	nif->rx_alloc = nbuf_rx_allocate;
	nif->tx_alloc = nbuf_tx_allocate;
	nif->rx_free = nbuf_rx_release;
	nif->tx_free = nbuf_tx_release;
	return TRUE;
}

/********************************************************************/
static int
lan91c111_reset (NIF *nif)
{
	int i;
	(void)nif;

	/* This resets the registers mostly to defaults, but doesn't
	   affect EEPROM.  That seems unnecessary */
	SMC_SELECT_BANK(0);
	LAN91C111_RCR = LAN91C111_RCR_SOFTRST;

	/* Setup the Configuration Register */
	/* This is necessary because the CONFIG_REG is not affected */
	/* by a soft reset */

	SMC_SELECT_BANK(1);
	LAN91C111_CFGR |= LAN91C111_CFGR_DEFAULT;

	SMC_SELECT_BANK(0);

	/* this should pause enough for the chip to be happy */
	for( i=0 ; i < 10000; i++);

	/* Disable transmit and receive functionality */
	LAN91C111_RCR = LAN91C111_RCR_CLEAR;
	LAN91C111_TCR = LAN91C111_TCR_CLEAR;

	/* set the control register to automatically
	   release successfully transmitted packets, to make the best
	   use out of our limited memory */
	SMC_SELECT_BANK(1);
	LAN91C111_CTRLR |= LAN91C111_CTRLR_AUTO_RELEASE;

	/* Reset the MMU */
	SMC_SELECT_BANK(2);
	LAN91C111_MMCR = LAN91C111_MMCR_RESET;
	while (LAN91C111_MMCR & LAN91C111_MMCR_BUSY);

	/* Disable all interrupts */
	LAN91C111_IMR = 0;

	return TRUE;
}

/********************************************************************/
static void
lan91c111_start (NIF *nif)
{
	/* Configure the Phy */
	lan91c111_phy_configure();

	/* now, enable interrupts */
	SMC_SELECT_BANK(2);
	LAN91C111_IMR = (LAN91C111_IMR_RCV_INT | LAN91C111_IMR_TX_INT);

	SMC_SELECT_BANK(0);
	/* see the header file for options in TCR/RCR DEFAULT*/

	/* enable transmit and recieve */
	LAN91C111_RCR = LAN91C111_RCR_DEFAULT;
	LAN91C111_TCR = LAN91C111_TCR_DEFAULT;

	/* Set Hardware Address */
	SMC_SELECT_BANK(1);
	LAN91C111_IAR01	= (uint16)(((nif->hwa[0])<<8) | nif->hwa[1]);
	LAN91C111_IAR23	= (uint16)(((nif->hwa[2])<<8) | nif->hwa[3]);
	LAN91C111_IAR45	= (uint16)(((nif->hwa[4])<<8) | nif->hwa[5]);
}

/********************************************************************/
static void
lan91c111_stop (NIF *nif)
{
	(void)nif;
	
	SMC_SELECT_BANK(0);

	/* Disable transmit and receive functionality */
	LAN91C111_RCR = LAN91C111_RCR_CLEAR;
	LAN91C111_TCR = LAN91C111_TCR_CLEAR;

	/* Shutdown Phy */
	lan91c111_phy_shutdown();
}

/********************************************************************/
static int
lan91c111_send (NIF *nif,
	HWA_ADDR_P	eth_dest,
	HWA_ADDR_P	eth_src,
	uint16		eth_type,
	NBUF		*pNbuf)
{
	/*
	 * This routine transmits one frame.  This routine only accepts
	 * 6-byte Ethernet addresses.
	 */
	uint32 mod_data_length, index, success, timeout, j;
	uint16 numPages, isr;

	/*
	 * The format of an Ethernet frame looks like:
	 *
	 *   00 01 02 03 04 05
	 *  +--+--+--+--+--+--+
	 *  |  |  |  |  |  |  | destination ethernet address
	 *  +--+--+--+--+--+--+
	 *   06 07 08 09 0A 0B
	 *  +--+--+--+--+--+--+
	 *  |  |  |  |  |  |  | source ethernet address
	 *  +--+--+--+--+--+--+
	 *   0C 0D
	 *  +--+--+
	 *  |  |  | type (IP = 0x0800)
	 *  +--+--+
	 *   0E 0F ...
	 *  +--+--+--+--+--+--+
	 *  |  |  |  |  |  |  | data (min of 46 octets, max of 1500)
	 *  +--+--+--+--+--+--+
	 *
	 *  +--+--+--+--+
	 *  |  |  |  |  | 32-bit CRC (generated by hardware)
	 *  +--+--+--+--+
	 *
	 * NOTE:  The NS8390 takes care of pre-amble, and is programmed
	 * to generate the 32-bit CRC.
	 *
	 * NOTE:  The minimum size of data transmitted in an ethernet
	 * frame is 46 octets (bytes).
	 */

	/*
	 * Check for valid length of data.
	 */
	if ((pNbuf->length > nif->mtu) || (pNbuf->length <= 0))
		return 0;

	/* 
	 * Modify Data Length if necessary 
	 */
	mod_data_length = pNbuf->length;

	/* 
	 * Align data length to 16-bits 
	 */
	if (mod_data_length & 0x0001)
		++mod_data_length;

	/* 
	 * Verify data length >= minimum Ethernet packet size 
	 */
	if (mod_data_length < ETH_MIN_SIZE)
		mod_data_length = ETH_MIN_SIZE;

	/*
	 * During the course of loading a frame into the FIFO, we
	 * can not have interrupts occur, especially those from the
	 * receiver!
	 */
	board_irq_disable();

	/* 
	 * Allocate memory for TX 
	 */
	SMC_SELECT_BANK(2);
	numPages = (uint16)(mod_data_length >> 8);
	
	if (numPages > 7 )	/* Too big */
		return FALSE;

	LAN91C111_MMCR = (uint16)(LAN91C111_MMCR_ALLOC | (numPages <<8));
	while (LAN91C111_MMCR & LAN91C111_MMCR_BUSY)
		;
	
	/* 
	 * Wait for allocation to complete 
	 */
	timeout = 100; 

	do {
		isr = LAN91C111_ISR;
		if (isr & LAN91C111_ISR_ALLOC_INT)
			{
			// allocation complete
			LAN91C111_ISR = (isr & 0x00FF) | LAN91C111_ISR_ALLOC_INT;
			break;
			}

		/* delay some time */
		for( j=0 ; j<100 ; j++);

   	} while ( -- timeout );

	if (timeout < 1)
		{
		printf("SMSC MMU memory allocation timeout");
		return FALSE;
		}
	
	/* 
	 * Copy the TX packet number into the Packet Number Register 
	 */
	LAN91C111_PNR = (uint16)((LAN91C111_ARR & 0x3F)<<8);

	/* 
	 * Check to see if FIFO is emtpy 
	 */
	if(LAN91C111_PTR & LAN91C111_PTR_NOT_EMPTY)
	{
		return FALSE;
	}

	/* 
	 * Write the Pointer Register 
	 */
	LAN91C111_PTR = LAN91C111_PTR_AUTOINC;

	/* 
	 * Write status word  to Data area
	 */
	LAN91C111_DATA = 0x0000;

	/* 
	 * Write Byte count 
	 */
	LAN91C111_DATA = (uint16)UBLBSWTICH(mod_data_length + 20);

	/*
	 * Write destination ethernet address 
	 */
	LAN91C111_DATA = *((uint16 *)&eth_dest[0]);
	LAN91C111_DATA = *((uint16 *)&eth_dest[2]);
	LAN91C111_DATA = *((uint16 *)&eth_dest[4]);

	/* 
	 * Write source ethernet address 
	 */
	LAN91C111_DATA = *((uint16 *)&eth_src[0]);
	LAN91C111_DATA = *((uint16 *)&eth_src[2]);
	LAN91C111_DATA = *((uint16 *)&eth_src[4]);

	/* 
	 * Write type 
	 */
	LAN91C111_DATA = eth_type;

	/* 
	 * Write data 
	 */
	for (index = 0; (index + 1) < pNbuf->length; index += 2)
	{
		LAN91C111_DATA = *((uint16*)&pNbuf->data[ETH_HDR_SIZE +index]);
	}

	/* 
	 * Write odd byte and pad with an extra byte to maintain 16-bit alignment 
	 */
	if(pNbuf->length & 0x0001)
	{
		LAN91C111_DATA = (uint16)(pNbuf->data[ETH_HDR_SIZE + pNbuf->length - 1] << 8);
		++pNbuf->length;
	}

	/* 
	 * Pad with 0x00's if original size was less than ETH_MIN_SIZE 
	 */
	for(index=pNbuf->length; index < ETH_MIN_SIZE; index += 2)
		LAN91C111_DATA = 0;

	/* 
	 * Write control byte - already aligned to always be even 
	 */
	LAN91C111_DATA = 0x0010;

	/* 
	 * Issus "ENQUEUE PACKET NUMBER TO TX FIFO 
	 */
	LAN91C111_MMCR = LAN91C111_MMCR_ENQUEUE;
	while (LAN91C111_MMCR & LAN91C111_MMCR_BUSY)
		;

	/* 
	 * Free the network buffer 
	 */
	nif->tx_free(pNbuf);

	/* 
	 * Allow interrupts again 
	 */
	board_irq_enable();

	return TRUE;
}

/********************************************************************/
static void
lan91c111_receive (void *not_used, NIF *nif)
{
	uint16 pktnum;
	uint16 pktlength;
	uint16 status;
	uint16 lastword;
	uint32 index;
	eth_frame_hdr * ethframe;
	NBUF * pNbuf;

	(void)not_used;
	
	/* I should already be in BANK 2 */
	pktnum = LAN91C111_RXFIFOR;

	if ( pktnum & LAN91C111_RXFIFOR_REMPTY ) 
	{
		/* we got called , but nothing was on the FIFO */
		/* don't need to restore anything */
		return;
	}

	/* Check to see if FIFO is emtpy */
	if(LAN91C111_PTR & LAN91C111_PTR_NOT_EMPTY)
	{
		printf("FIFO is NOT EMPTY\n");
		return;
	}

	/* Write Addr Ptr */
	/* Write Address Pointer Reg */
	LAN91C111_PTR = (LAN91C111_PTR_AUTOINC | LAN91C111_PTR_RCV |LAN91C111_PTR_READ);

	/* Read Word 0 from RAM */

	/* First two words are status and packet_length */
	status 		= LAN91C111_DATA;
	pktlength 	= LAN91C111_DATA;
	pktlength 	= (uint16)(UBLBSWTICH(pktlength) & 0x07FF);

	/* Status work ok? */
	if(status & RS_ERRORS)
	{
		printf("error in receive status word: %04X\n",status);
	}

	/* check if multicast*/
	if((status & RS_MULTICAST) | (status & RS_BRODCAST))
	{	
		/* Issue "Remove and Release" command */
		while (LAN91C111_MMCR & LAN91C111_MMCR_BUSY);
		LAN91C111_MMCR = LAN91C111_MMCR_RELEASE;
		return;
	}

	/* Obtain system RAM buffer for the frame */
	if ((pNbuf = nif->rx_alloc()) == NULL)
	{
		printf("error, cannot obtain system RAM buffer for the frame\n");
		/* Issue "Remove and Release" command */		
		while (LAN91C111_MMCR & LAN91C111_MMCR_BUSY);
		LAN91C111_MMCR = LAN91C111_MMCR_RELEASE;
		return;
	}

	ethframe = (eth_frame_hdr *)&pNbuf->data[ETH_HDR_OFFSET];

	/* Copy Data Per upper Layer Specs */

	/* get destination */
	*(uint16*)&ethframe->dest[0] = LAN91C111_DATA;
	*(uint16*)&ethframe->dest[2] = LAN91C111_DATA;
	*(uint16*)&ethframe->dest[4] = LAN91C111_DATA;
	/* get source */
	*(uint16*)&ethframe->src[0] = LAN91C111_DATA;
	*(uint16*)&ethframe->src[2] = LAN91C111_DATA;
	*(uint16*)&ethframe->src[4] = LAN91C111_DATA;
	/* get frame type */
	ethframe->type = LAN91C111_DATA;

	/* check if we can handle this type */
	if(!nif_protocol_exist(nif, ethframe->type))
	{
		/* Issue "Remove and Release" command */
		printf("can't handle this frame\n");
		while (LAN91C111_MMCR & LAN91C111_MMCR_BUSY);
		LAN91C111_MMCR = LAN91C111_MMCR_RELEASE;

		nif->rx_free(pNbuf);
		return;
	}

	/* Get length of data in frame */
	/* 6 for des address  */
	/* 6 for src address  */
	/* 2 for type         */
	/* 6 for SMSC frame (status word, byte count, control byte and odd byte) */

	pNbuf->length = pktlength - (6+6+2+6);

	for(index=0 ; index < pNbuf->length; index +=2)
	{
		*(uint16*)&ethframe->data[index] = LAN91C111_DATA;
	}

	/* Get control byte and odd byte if there is one */
	lastword = LAN91C111_DATA;
	
	if(lastword & 0x0020)
	{
		/* write last byte */
		ethframe->data[index] = (uint8)((lastword & 0xFF00)>>8);
		/* increase length by 1*/
		pNbuf->length++;
	}

	/* Issue "Remove and Release" command */
	while (LAN91C111_MMCR & LAN91C111_MMCR_BUSY);
	LAN91C111_MMCR = LAN91C111_MMCR_RELEASE;

	/* Call the appropriate handler */
	nif_protocol_handler(nif,ethframe->type,pNbuf);

	/* free the buffer */
	nif->rx_free(pNbuf);

	/* Return to ISR */
	return;
}

/********************************************************************/
int
lan91c111_handler (void * applefritter, NIF *nif)
{
	uint16 intstatus, oldIMR, oldBank, oldPTR;
	vuint32 intdone;

	intdone = FALSE;

	/* Save off Bank Select register */
	oldBank = LAN91C111_BSR;

	/* switch to Bank 2 */
	SMC_SELECT_BANK(2);

	/* Save off Address Pointer register */
	oldPTR = LAN91C111_PTR;

	/* read the interrupt mask register */
	oldIMR = LAN91C111_IMR;

	/* disable all interrupts */
	LAN91C111_IMR = 0;

	while(!intdone)
	{
		/* switch to Bank 2 */
		SMC_SELECT_BANK(2);

		/* read in interrupt status register */
		intstatus = (LAN91C111_ISR&0xFF00);

		if(LAN91C111_ISR_RCV_INT &intstatus)		
		{
			lan91c111_receive(applefritter, nif);
			++nif->f_rx;

		}
		else if(LAN91C111_ISR_TX_INT &intstatus)		
		{
			/* Acknowledge TXINTR */
			LAN91C111_ISR = LAN91C111_ISR_TX_INT;
			++nif->f_tx;
		}
		else if(LAN91C111_ISR_RX_OVRN_INT &intstatus)	
		{
			/* Acknowledge interrupt */
			LAN91C111_ISR = LAN91C111_ISR_RX_OVRN_INT;
		}

		else
			intdone = TRUE;

	} /* while(!intdone) */
	
	/* Make sure I am in Bank 2 */
	SMC_SELECT_BANK(2);
	
	/* Check to see if FIFO is emtpy */
	if(LAN91C111_PTR & LAN91C111_PTR_NOT_EMPTY)
	{
		return FALSE;
	}

	/* Restore Address Pointer register */
	LAN91C111_PTR = oldPTR; 

	
	/* Restore the interrupt mask register */
	LAN91C111_IMR = oldIMR;

	/* Restore Bank Select register */
	LAN91C111_BSR = oldBank;

	return TRUE;
}
/********************************************************************
 * Shutdown phy
 ********************************************************************/
static void lan91c111_phy_shutdown(void)
{
	unsigned int timeout;
	unsigned int j;

	// Reset the PHY, setting all other bits to zero
	lan91c111_write_phy_reg(PHYADDR, PHY_CNTL_REG, PHY_CNTL_RST);

	// Wait for the reset to complete, or time out
	timeout = 6; // Wait up to 3 seconds
	while (timeout--)
		{
		if (!(lan91c111_read_phy_reg(PHYADDR, PHY_CNTL_REG) & PHY_CNTL_RST))
			{
			// reset complete
			break;
			}

		/* delay some time */
		for( j=0 ; j<100 ; j++);
		}

	if (timeout < 1)
		{
		printf("PHY reset timed out\n");
		return;
		}
}

/********************************************************************
 * Configures the specified PHY using Autonegotiation. Calls
 * smc_phy_fixed() if the user has requested a certain config.
 ********************************************************************/
static void lan91c111_phy_configure(void)
{
	unsigned int timeout;
	unsigned int j;
	uint16		 status;

	// Reset the PHY, setting all other bits to zero
	lan91c111_write_phy_reg(PHYADDR, PHY_CNTL_REG, PHY_CNTL_RST);

	// Wait for the reset to complete, or time out
	timeout = 6; // Wait up to 3 seconds
	while (timeout--)
		{
		if (!(lan91c111_read_phy_reg(PHYADDR, PHY_CNTL_REG) & PHY_CNTL_RST))
			{
			// reset complete
			break;
			}

		/* delay some time */
		for( j=0 ; j<100 ; j++);
		}

	if (timeout < 1)
		{
		printf("PHY reset timed out\n");
		return;
		}

	// Enable PHY Interrupts (for register 18)
	// Interrupts listed here are disabled
	lan91c111_write_phy_reg(PHYADDR, PHY_MASK_REG, PHY_INT_LNKFAIL
												|PHY_INT_INT
												|PHY_INT_LOSSSYNC
												|PHY_INT_CWRD
												|PHY_INT_SSD
												|PHY_INT_ESD	
												|PHY_INT_RPOL
												|PHY_INT_JAB	
												|PHY_INT_SPDDET
												|PHY_INT_DPLXDET);

	/* Configure the Receive/Phy Control register */
	SMC_SELECT_BANK(0);
	LAN91C111_RPC = LAN91C111_RPC_DEFAULT;


	// Update our Auto-Neg Advertisement Register
	lan91c111_write_phy_reg(PHYADDR, PHY_AD_REG, (PHY_AD_TX_FDX|PHY_AD_TX_HDX|PHY_AD_10_FDX|PHY_AD_10_HDX|PHY_AD_CSMA));		

	// Restart auto-negotiation process in order to advertise my caps
	lan91c111_write_phy_reg(PHYADDR, PHY_CNTL_REG,	PHY_CNTL_ANEG_EN | PHY_CNTL_ANEG_RST );
	
	// Wait for the auto-negotiation to complete.  This may take from
	// 2 to 3 seconds.
	// Wait for the reset to complete, or time out
	timeout = 20; // Wait up to 10 seconds
	while (timeout--)
	{
		status = lan91c111_read_phy_reg(PHYADDR, PHY_STAT_REG);
		if (status & PHY_STAT_ANEG_ACK)
		{
			// auto-negotiate complete
			break;
		}

		/* delay some time */
		for( j=0 ; j<10000 ; j++);

		// Restart auto-negotiation if remote fault
		if (status & PHY_STAT_REM_FLT)
		{
			// Restart auto-negotiation
			lan91c111_write_phy_reg(PHYADDR, PHY_CNTL_REG,
				PHY_CNTL_ANEG_EN | PHY_CNTL_ANEG_RST |
				PHY_CNTL_SPEED | PHY_CNTL_DPLX);
		}
	}
}

/********************************************************************
 * Writes a register to the MII Management serial interface
 ********************************************************************/
static void lan91c111_write_phy_reg(uint8 phyaddr, uint8 phyreg, uint16 phydata)
{
	uint16 oldBank;
	uint32 i,j;
	uint16 mask;
	uint16 mii_reg;
	uint16 bits[65];
	uint32 clk_idx = 0;

	// 32 consecutive ones on MDO to establish sync
	for (i = 0; i < 32; ++i)
		bits[clk_idx++] = LAN91C111_MII_MDOE | LAN91C111_MII_MDO;

	// Start code <01>
	bits[clk_idx++] = LAN91C111_MII_MDOE;
	bits[clk_idx++] = LAN91C111_MII_MDOE | LAN91C111_MII_MDO;

	// Write command <01>
	bits[clk_idx++] = LAN91C111_MII_MDOE;
	bits[clk_idx++] = LAN91C111_MII_MDOE | LAN91C111_MII_MDO;

	// Output the PHY address, msb first
	mask = (uint8)0x10;
	for (i = 0; i < 5; ++i)
		{
		if (phyaddr & mask)
			bits[clk_idx++] = LAN91C111_MII_MDOE | LAN91C111_MII_MDO;
		else
			bits[clk_idx++] = LAN91C111_MII_MDOE;

		// Shift to next lowest bit
		mask >>= 1;
		}

	// Output the phy register number, msb first
	mask = (uint8)0x10;
	for (i = 0; i < 5; ++i)
		{
		if (phyreg & mask)
			bits[clk_idx++] = LAN91C111_MII_MDOE | LAN91C111_MII_MDO;
		else
			bits[clk_idx++] = LAN91C111_MII_MDOE;

		// Shift to next lowest bit
		mask >>= 1;
		}

	// Tristate and turnaround (2 bit times)
	bits[clk_idx++] = 0;
	bits[clk_idx++] = 0;

	// Write out 16 bits of data, msb first
	mask = 0x8000;
	for (i = 0; i < 16; ++i)
		{
		if (phydata & mask)
			bits[clk_idx++] = LAN91C111_MII_MDOE | LAN91C111_MII_MDO;
		else
			bits[clk_idx++] = LAN91C111_MII_MDOE;

		// Shift to next lowest bit
		mask >>= 1;
		}

	// Save the current bank
	oldBank = LAN91C111_BSR;

	// Select bank 3
	SMC_SELECT_BANK(3);

	// Get the current MII register value
	mii_reg = LAN91C111_MII;

	// Turn off all MII Interface bits
	mii_reg &= ~(LAN91C111_MII_MDOE|LAN91C111_MII_MCLK|LAN91C111_MII_MDI|LAN91C111_MII_MDO);

	// Clock all cycles
	for (i = 0; i < sizeof bits; ++i)
		{
		// Clock Low - output data
		LAN91C111_MII = (mii_reg | bits[i]);
		for(j=0 ; j<4000 ;j++);
		
		// Clock Hi - input data
		LAN91C111_MII = (mii_reg | bits[i] | LAN91C111_MII_MCLK);
		for(j=0 ; j<4000 ;j++);

		bits[i] |= 	(LAN91C111_MII & LAN91C111_MII_MDI);		
		}

	// Return to idle state
	// Set clock to low, data to low, and output tristated
	LAN91C111_MII = mii_reg;
	for(j=0 ; j<1000 ;j++);

	// Restore original bank select
//	SMC_SELECT_BANK(oldBank);
	LAN91C111_BSR = oldBank;

}

/********************************************************************
 * Reads a register from the MII Management serial interface
 ********************************************************************/
static uint16
lan91c111_read_phy_reg(uint8 phyaddr, uint8 phyreg)
{
	uint16 oldBank;
	uint32 i,j;
	uint8 mask;
	uint16 mii_reg;
	uint16 bits[64];
	uint32 clk_idx = 0;
	uint32 input_idx;
	uint16 phydata;

	// 32 consecutive ones on MDO to establish sync
	for (i = 0; i < 32; ++i)
		bits[clk_idx++] = LAN91C111_MII_MDOE | LAN91C111_MII_MDO;

	// Start code <01>
	bits[clk_idx++] = LAN91C111_MII_MDOE;
	bits[clk_idx++] = LAN91C111_MII_MDOE | LAN91C111_MII_MDO;

	// Read command <10>
	bits[clk_idx++] = LAN91C111_MII_MDOE | LAN91C111_MII_MDO;
	bits[clk_idx++] = LAN91C111_MII_MDOE;

	// Output the PHY address, msb first
	mask = (uint8)0x10;
	for (i = 0; i < 5; ++i)
		{
		if (phyaddr & mask)
			bits[clk_idx++] = LAN91C111_MII_MDOE | LAN91C111_MII_MDO;
		else
			bits[clk_idx++] = LAN91C111_MII_MDOE;

		// Shift to next lowest bit
		mask >>= 1;
		}

	// Output the phy register number, msb first
	mask = (uint8)0x10;
	for (i = 0; i < 5; ++i)
		{
		if (phyreg & mask)
			bits[clk_idx++] = LAN91C111_MII_MDOE | LAN91C111_MII_MDO;
		else
			bits[clk_idx++] = LAN91C111_MII_MDOE;

		// Shift to next lowest bit
		mask >>= 1;
		}

	// Tristate and turnaround (2 bit times)
	bits[clk_idx++] = 0;
	//bits[clk_idx++] = 0;

	// Input starts at this bit time
	input_idx = clk_idx;

	// Will input 16 bits
	for (i = 0; i < 16; ++i)
		bits[clk_idx++] = 0;

	// Final clock bit
	bits[clk_idx] = 0;

	// Save the current bank
	oldBank = LAN91C111_BSR;

	// Select bank 3
	SMC_SELECT_BANK(3);

	// Get the current MII register value
	mii_reg = LAN91C111_MII;

	// Turn off all MII Interface bits
	mii_reg &= ~(LAN91C111_MII_MDOE|LAN91C111_MII_MCLK|LAN91C111_MII_MDI|LAN91C111_MII_MDO);

	// Clock all 64 cycles
	for (i = 0; i < sizeof bits; ++i)
		{
		// Clock Low - output data
		LAN91C111_MII = ( mii_reg | bits[i] );
		for(j=0 ; j<4000 ;j++);

		// Clock Hi - input data
		LAN91C111_MII = ( mii_reg | bits[i] | LAN91C111_MII_MCLK );
		for(j=0 ; j<4000 ;j++)
		{}
		bits[i] |= (LAN91C111_MII & LAN91C111_MII_MDI);
		}

	// Return to idle state
	// Set clock to low, data to low, and output tristated
	LAN91C111_MII = mii_reg;
	for(j=0 ; j<1000 ;j++);

	// Restore original bank select
//	SMC_SELECT_BANK( oldBank );
	LAN91C111_BSR = oldBank;

	// Recover input data
	phydata = 0;
	for (i = 0; i < 16; ++i)
		{
		phydata <<= 1;

		if (bits[input_idx++] & LAN91C111_MII_MDI)
			phydata |= 0x0001;
		}

	return(phydata);	
}

/********************************************************************/
